#pragma once

#include "Config.h"

#include <complex> // Must be included before fftw3
#include <functional>
#include <memory>
#include <vector>

#include <fftw3.h>
#include <yaml-cpp/yaml.h>

#include "QGInputDevice.h"

#ifdef HAVE_LIBLIQUIDSDR
#include <liquid/liquid.h>
#else
#ifdef HAVE_LIBRTFILTER
#include <rtfilter.h>
#endif // HAVE_LIBRTFILTER
#endif // HAVE_LIBLIQUIDSDR

class QGProcessor {
public:
	QGProcessor(const YAML::Node &config, const QGInputDevice& inputDevice);
	~QGProcessor();

	void addCb(std::function<void(const std::complex<float>*)>cb);

	void addIQ(const std::complex<float> *iq);

	unsigned int sampleRate() const { return _sampleRate; };
	unsigned int chunkSize() const { return _chunkSize; };
	int fftSize() const { return _N; };
	int fftOverlap() const { return _overlap; };
	float downsamplingRate() const { return _rate; };

	const QGInputDevice &inputDevice;

private:
	unsigned int _resample(const std::complex<float> *in, std::complex<float> *out);
	void _fft();

	unsigned int _sampleRate;
	unsigned int _chunkSize;
	int _N;
	int _overlap; // 0: no overlap, 1: 1/2, 2: 2/3...
	float _rate;

	bool _noProcessor;

	// Input buffer
	int _inputIndex;
	std::unique_ptr<std::complex<float>[]> _input;

	// Window function
	std::unique_ptr<float[]> _hannW;

	// Resampler
#ifdef HAVE_LIBLIQUIDSDR
	resamp_crcf _liquidSdrResampler;
#else
#ifdef HAVE_LIBRTFILTER
	hfilter _rtFilterResampler;
#else
	unsigned int _counter;
#endif // HAVE_LIBRTFILTER
#endif // HAVE_LIBLIQUIDSDR

	// FFT
	std::complex<float> *_fftIn;
	std::complex<float> *_fftOut;
	fftwf_plan _plan;

	std::vector<std::function<void(const std::complex<float>*)>> _cbs;
};
