#include "catch_v2.2.3/catch.hpp"

#include <limits>

#include "QGInputDevice.h"
#include "QGPlugin.h"
#include "QGProcessor.h"
#include "QGUtils.h"


TEST_CASE("QGInputDevice defaults", "[input]") {
  QGInputDevice device(YAML::Load("{}"));
  CHECK(device.type() == "");
  CHECK(device.sampleRate() == 48000);
  CHECK(device.dispFreq() == Approx(0).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device.tunerFreq() == Approx(0).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device.centerFreq() == Approx(0.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device.ppm() == Approx(0.).epsilon(std::numeric_limits<double>::epsilon()*100));
}

TEST_CASE("QGInputDevice freq params", "[input]") {
  QGInputDevice device1(YAML::Load("{samplerate: 96000, basefreq: 7000000}"));
  CHECK(device1.type() == "");
  CHECK(device1.sampleRate() == 96000);
  CHECK(device1.dispFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device1.tunerFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device1.centerFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device1.ppm() == Approx(0.).epsilon(std::numeric_limits<double>::epsilon()*100));

  QGInputDevice device2(YAML::Load("{samplerate: 96000, basefreq: 7000000, ppm: -5}"));
  CHECK(device2.type() == "");
  CHECK(device2.sampleRate() == 96000);
  CHECK(device2.dispFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device2.tunerFreq() == Approx(7000035.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device2.centerFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device2.ppm() == Approx(-5.).epsilon(std::numeric_limits<double>::epsilon()*100));

  QGInputDevice device3(YAML::Load("{samplerate: 96000, basefreq: 7000000, ppm: 2}"));
  CHECK(device3.type() == "");
  CHECK(device3.sampleRate() == 96000);
  CHECK(device3.dispFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device3.tunerFreq() == Approx(6999986.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device3.centerFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device3.ppm() == Approx(2.).epsilon(std::numeric_limits<double>::epsilon()*100));
}

TEST_CASE("QGInputDevice offset freq params", "[input]") {
  QGInputDevice device1(YAML::Load("{samplerate: 96000, basefreq: 7000000, offset: 100000000}"));
  CHECK(device1.type() == "");
  CHECK(device1.sampleRate() == 96000);
  CHECK(device1.dispFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device1.tunerFreq() == Approx(107000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device1.centerFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device1.ppm() == Approx(0.).epsilon(std::numeric_limits<double>::epsilon()*100));

  QGInputDevice device2(YAML::Load("{samplerate: 96000, basefreq: 7000000, offset: 100000000, offsetppm: -2}"));
  CHECK(device2.type() == "");
  CHECK(device2.sampleRate() == 96000);
  CHECK(device2.dispFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device2.tunerFreq() == Approx(106999800.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device2.centerFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device2.ppm() == Approx(0.).epsilon(std::numeric_limits<double>::epsilon()*100));

  QGInputDevice device3(YAML::Load("{samplerate: 96000, basefreq: 7000000, offset: 100000000, offsetppm: 1}"));
  CHECK(device3.type() == "");
  CHECK(device3.sampleRate() == 96000);
  CHECK(device3.dispFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device3.tunerFreq() == Approx(107000100.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device3.centerFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device3.ppm() == Approx(0.).epsilon(std::numeric_limits<double>::epsilon()*100));

  QGInputDevice device4(YAML::Load("{samplerate: 96000, basefreq: 7000000, ppm: 2, offset: 100000000, offsetppm: -1}"));
  CHECK(device4.type() == "");
  CHECK(device4.sampleRate() == 96000);
  CHECK(device4.dispFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device4.tunerFreq() == Approx(106999686.0002).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device4.centerFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device4.ppm() == Approx(2.).epsilon(std::numeric_limits<double>::epsilon()*100));

  QGInputDevice device5(YAML::Load("{samplerate: 96000, basefreq: 7000000, ppm: -1, offset: 100000000, offsetppm: 2}"));
  CHECK(device5.type() == "");
  CHECK(device5.sampleRate() == 96000);
  CHECK(device5.dispFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device5.tunerFreq() == Approx(107000307.0002).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device5.centerFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
  CHECK(device5.ppm() == Approx(-1.).epsilon(std::numeric_limits<double>::epsilon()*100));
}

TEST_CASE("QGInputAlsa", "[input]") {
  QGPlugin<QGInputDevice>* plugin = new QGPlugin<QGInputDevice>("input");
  if (plugin->has_dl("alsa")) {
    std::unique_ptr<QGInputDevice> device;

    // Check type, samplerate and freq/ppm
    REQUIRE_NOTHROW(device = QGInputDevice::CreateInputDevice(YAML::Load("{type: alsa, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5}")));
    CHECK(device->type() == "Alsa");
    CHECK(device->sampleRate() == 96000);
    CHECK(device->dispFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->tunerFreq() == Approx(0.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->centerFreq() == Approx(6999965.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->ppm() == Approx(-5.).epsilon(std::numeric_limits<double>::epsilon()*100));

    // Check different channel type are accepted
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: alsa, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, channel: mono}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: alsa, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, channel: left}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: alsa, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, channel: right}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: alsa, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, channel: iq}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: alsa, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, channel: inviq}")));
    CHECK_THROWS(QGInputDevice::CreateInputDevice(YAML::Load("{type: alsa, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, channel: stereo}")));
  }
  delete plugin;
}

TEST_CASE("QGInputHackRF", "[input]") {
  QGPlugin<QGInputDevice>* plugin = new QGPlugin<QGInputDevice>("input");
  if (plugin->has_dl("hackrf")) {
    std::unique_ptr<QGInputDevice> device;

    // Check type, samplerate and freq/ppm
    REQUIRE_NOTHROW(device = QGInputDevice::CreateInputDevice(YAML::Load("{type: hackrf, nodevice: true, samplerate: 8000000, basefreq: 7000000, ppm: -5}")));
    CHECK(device->type() == "HackRF");
    CHECK(device->sampleRate() == 8000000);
    CHECK(device->dispFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->tunerFreq() == Approx(7000035.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->centerFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->ppm() == Approx(-5.).epsilon(std::numeric_limits<double>::epsilon()*100));

    // Check device serial and index are accepted but not together
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: hackrf, deviceserial: 4242, nodevice: true}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: hackrf, deviceindex: 1, nodevice: true}")));
    CHECK_THROWS(QGInputDevice::CreateInputDevice(YAML::Load("{type: hackrf, deviceserial: 4242, deviceindex: 1, nodevice: true}")));
  }
  delete plugin;
}

TEST_CASE("QGInputLime", "[input]") {
  QGPlugin<QGInputDevice>* plugin = new QGPlugin<QGInputDevice>("input");
  if (plugin->has_dl("limesdr")) {
    std::unique_ptr<QGInputDevice> device;

    // Check type, samplerate and freq/ppm
    REQUIRE_NOTHROW(device = QGInputDevice::CreateInputDevice(YAML::Load("{type: limesdr, nodevice: true, samplerate: 8000000, basefreq: 7000000, ppm: -5}")));
    CHECK(device->type() == "LimeSdr");
    CHECK(device->sampleRate() == 8000000);
    CHECK(device->dispFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->tunerFreq() == Approx(7000035.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->centerFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->ppm() == Approx(-5.).epsilon(std::numeric_limits<double>::epsilon()*100));
  }
  delete plugin;
}

TEST_CASE("QGInputRtlSdr", "[input]") {
  QGPlugin<QGInputDevice>* plugin = new QGPlugin<QGInputDevice>("input");
  if (plugin->has_dl("rtlsdr")) {
    std::unique_ptr<QGInputDevice> device;

    // Check type, samplerate and freq/ppm
    REQUIRE_NOTHROW(device = QGInputDevice::CreateInputDevice(YAML::Load("{type: rtlsdr, nodevice: true, samplerate: 240000, basefreq: 7000000, ppm: -5.2}")));
    CHECK(device->type() == "RtlSdr");
    CHECK(device->sampleRate() == 240000);
    CHECK(device->dispFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->tunerFreq() == Approx(7000001.4).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->centerFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->ppm() == Approx(-.2).epsilon(std::numeric_limits<double>::epsilon()*100));

    // Check different settings for directsampling are accepted
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: rtlsdr, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, directsampling: false}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: rtlsdr, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, directsampling: true}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: rtlsdr, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, directsampling: true, directsamplingbranch: i-branch}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: rtlsdr, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, directsampling: true, directsamplingbranch: q-branch}")));
    CHECK_THROWS(QGInputDevice::CreateInputDevice(YAML::Load("{type: rtlsdr, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, directsampling: maybe}")));
    CHECK_THROWS(QGInputDevice::CreateInputDevice(YAML::Load("{type: rtlsdr, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, directsampling: true, directsamplingbranch: y-branch}")));
  }
  delete plugin;
}

TEST_CASE("QGInputStdin", "[input]") {
  QGPlugin<QGInputDevice>* plugin = new QGPlugin<QGInputDevice>("input");
  if (plugin->has_dl("stdin")) {
    std::unique_ptr<QGInputDevice> device;

    // Check type, samplerate and freq/ppm
    REQUIRE_NOTHROW(device = QGInputDevice::CreateInputDevice(YAML::Load("{type: stdin, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5}")));
    CHECK(device->type() == "Stdin");
    CHECK(device->sampleRate() == 96000);
    CHECK(device->dispFreq() == Approx(7000000.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->tunerFreq() == Approx(0.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->centerFreq() == Approx(6999965.).epsilon(std::numeric_limits<double>::epsilon()*100));
    CHECK(device->ppm() == Approx(-5.).epsilon(std::numeric_limits<double>::epsilon()*100));

    // Check different format type are accepted
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: stdin, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, format: u8iq}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: stdin, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, format: s8iq}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: stdin, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, format: u16iq}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: stdin, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, format: s16iq}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: stdin, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, format: s16mono}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: stdin, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, format: s16left}")));
    CHECK_NOTHROW(QGInputDevice::CreateInputDevice(YAML::Load("{type: stdin, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, format: s16right}")));
    CHECK_THROWS(QGInputDevice::CreateInputDevice(YAML::Load("{type: stdin, nodevice: true, samplerate: 96000, basefreq: 7000000, ppm: -5, format: straight}")));
  }
  delete plugin;
}

TEST_CASE("QGPlugin", "[plugin]") {
  CHECK(QGPlugin<QGInputDevice>::plugin_path() == std::string(INSTALL_PREFIX) + "/lib/qrsspig/");
}

TEST_CASE("QGProcessor", "[processor]") {
  QGPlugin<QGInputDevice>* plugin = new QGPlugin<QGInputDevice>("input");

  std::unique_ptr<QGInputDevice> device;
  REQUIRE_NOTHROW(
    device = QGInputDevice::CreateInputDevice(
      YAML::Load(
        "{type: stdin, nodevice: true, samplerate: 16000, basefreq: 7000000}"
      )
    )
  );

  // Check defaults
  std::unique_ptr<QGProcessor> processor;
  CHECK_NOTHROW(
    processor.reset(
      new QGProcessor(
        YAML::Load("{noprocessor: true}"),
        *(device.get())
      )
    )
  );

  CHECK(processor->sampleRate() == 16000);
  CHECK(processor->chunkSize() == 32);
  CHECK(processor->fftSize() == 2048);
  CHECK(processor->fftOverlap() == 1536);
  CHECK(processor->downsamplingRate() == 1.);

  delete plugin;
}

TEST_CASE("QGUtils formatFilename", "[utils]") {
  CHECK(QGUtils::formatFilename("", 14000000, std::chrono::milliseconds(0)) == "");
  CHECK(QGUtils::formatFilename("%f", 14000000, std::chrono::milliseconds(0)) == "14000000");
  CHECK(QGUtils::formatFilename("%D", 14000000, std::chrono::milliseconds(0)) == "1970-01-01");
  CHECK(QGUtils::formatFilename("%d", 14000000, std::chrono::milliseconds(0)) == "19700101");
  CHECK(QGUtils::formatFilename("%T", 14000000, std::chrono::milliseconds(0)) == "00:00:00");
  CHECK(QGUtils::formatFilename("%t", 14000000, std::chrono::milliseconds(0)) == "000000");
  CHECK(QGUtils::formatFilename("%I", 14000000, std::chrono::milliseconds(0)) == "1970-01-01T00:00:00Z");
  CHECK(QGUtils::formatFilename("%i", 14000000, std::chrono::milliseconds(0)) == "19700101T000000Z");
  CHECK(QGUtils::formatFilename("%%", 14000000, std::chrono::milliseconds(0)) == "%");
  CHECK(QGUtils::formatFilename("test_%fHz_%I", 14000000, std::chrono::milliseconds(0)) == "test_14000000Hz_1970-01-01T00:00:00Z");
  CHECK(QGUtils::formatFilename("test_%fHz_%i", 14000000, std::chrono::milliseconds(0)) == "test_14000000Hz_19700101T000000Z");
}

TEST_CASE("QGUtils levelBar", "[utils]") {
  CHECK(QGUtils::levelBar(std::numeric_limits<float>::infinity()) == "██████████████████████████████████████████████████");
  CHECK(QGUtils::levelBar(1000.)  == "██████████████████████████████████████████████████");
  CHECK(QGUtils::levelBar(0.)     == "██████████████████████████████████████████████████");
  CHECK(QGUtils::levelBar(-1.)    == "█████████████████████████████████████████████████▌");
  CHECK(QGUtils::levelBar(-2.)    == "█████████████████████████████████████████████████ ");
  CHECK(QGUtils::levelBar(-4.)    == "████████████████████████████████████████████████  ");
  CHECK(QGUtils::levelBar(-6.)    == "███████████████████████████████████████████████   ");
  CHECK(QGUtils::levelBar(-8.)    == "██████████████████████████████████████████████    ");
  CHECK(QGUtils::levelBar(-10.)   == "█████████████████████████████████████████████     ");
  CHECK(QGUtils::levelBar(-20.)   == "████████████████████████████████████████          ");
  CHECK(QGUtils::levelBar(-30.)   == "███████████████████████████████████               ");
  CHECK(QGUtils::levelBar(-40.)   == "██████████████████████████████                    ");
  CHECK(QGUtils::levelBar(-50.)   == "█████████████████████████                         ");
  CHECK(QGUtils::levelBar(-60.)   == "████████████████████                              ");
  CHECK(QGUtils::levelBar(-70.)   == "███████████████                                   ");
  CHECK(QGUtils::levelBar(-80.)   == "██████████                                        ");
  CHECK(QGUtils::levelBar(-90.)   == "█████                                             ");
  CHECK(QGUtils::levelBar(-100.)  == "                                                  ");
  CHECK(QGUtils::levelBar(-1000.) == "                                                  ");
  CHECK(QGUtils::levelBar(-std::numeric_limits<float>::infinity()) == "                                                  ");
}
